"""
Copyright (c) 2024, Auorui.
All rights reserved.
time 2024-01-25
"""
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from scipy.ndimage import distance_transform_edt
from skimage.segmentation import find_boundaries

class L1Loss(nn.Module):
    """
    L1 Loss (MAE Loss)
        Measure the average absolute difference between each element in the predicted output
    and the corresponding element in the target.
    Args:
        input (torch.Tensor): The predicted output.
        target (torch.Tensor): The target or ground truth.
    Returns:
        torch.Tensor: The L1 loss between input and target.
    """
    def __init__(self):
        super(L1Loss, self).__init__()

    def forward(self, y_pred, y_true):
        loss = torch.mean(torch.abs(y_pred - y_true))
        return loss

class L2Loss(nn.Module):
    """
    L2 Loss (MSE Loss)
        Measure the average squared difference between each element in the predicted
    output and the corresponding element in the target。

    Args:
        input (torch.Tensor): The predicted output.
        target (torch.Tensor): The target or ground truth.

    Returns:
        torch.Tensor: The L2 loss between input and target.
    """
    def __init__(self, eps=1e-3):
        super(L2Loss, self).__init__()
        self.esp = eps

    def forward(self, y_pred, y_true):
        loss = torch.mean(torch.pow(y_pred - y_true, 2) + self.esp)
        return loss

class BCELoss(nn.Module):
    """
    BCELoss (Binary Cross Entropy Loss)
        Used to measure the logarithmic probability difference between each element in
    the predicted output and the corresponding element in the target, implemented according to the formula.

    Args:
        input (torch.Tensor): The predicted output.Map to (0,1) through sigmoid function.
        target (torch.Tensor): The target or ground truth.

    Returns:
        torch.Tensor: The binary cross entropy loss between input and target.
    """
    def __init__(self):
        super(BCELoss, self).__init__()

    def forward(self, y_pred, y_true):
        y_pred = torch.sigmoid(y_pred)
        loss = - (y_true * torch.log(y_pred) + (1 - y_true) * torch.log(1 - y_pred))
        return loss.mean()


class MCCLoss(nn.Module):
    def __init__(self, eps: float = 1e-5, threshold: float = .5):
        """Compute Matthews Correlation Coefficient Loss for image segmentation task.
        It only supports binary mode.

        Args:
            eps (float): Small epsilon to handle situations where all the samples in the dataset belong to one class

        Reference:
            https://github.com/kakumarabhishek/MCC-Loss
        """
        super().__init__()
        self.eps = eps
        self.threshold = threshold

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        """Compute MCC loss
        MCC = (TP * TN - FP * FN) / sqrt((TP+FP) * (TP+FN) * (TN+FP) * (TN+FN))
        Args:
            y_pred (torch.Tensor): model prediction of shape (N, H, W) or (N, 1, H, W)
            y_true (torch.Tensor): ground truth labels of shape (N, H, W) or (N, 1, H, W)

        Returns:
            torch.Tensor: loss value (1 - mcc)
        """
        y_pred = (y_pred >= self.threshold).float()

        y_true = y_true.view(y_true.shape[0], 1, -1)
        y_pred = y_pred.view(y_true.shape[0], 1, -1)

        tp = torch.sum(torch.mul(y_pred, y_true)) + self.eps
        tn = torch.sum(torch.mul((1 - y_pred), (1 - y_true))) + self.eps
        fp = torch.sum(torch.mul(y_pred, (1 - y_true))) + self.eps
        fn = torch.sum(torch.mul((1 - y_pred), y_true)) + self.eps

        numerator = torch.mul(tp, tn) - torch.mul(fp, fn)
        denominator = torch.sqrt(torch.add(tp, fp) * torch.add(tp, fn) * torch.add(tn, fp) * torch.add(tn, fn))

        mcc = torch.div(numerator.sum(), denominator.sum())
        loss = 1.0 - mcc
        return loss

class CrossEntropyLoss(nn.Module):
    """
    Cross Entropy Loss:
        Used to measure the cross entropy between the predicted output and the target
    distribution.
    CE Loss = nn.NLLLoss(reduction=self.reduction)(F.log_softmax(y_pred, dim=1), y_true)

    Args:
        input (torch.Tensor): The predicted output (logits).
        target (torch.Tensor): The target or ground truth (class labels).
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.

    Returns:
        torch.Tensor: The cross entropy loss between input and target.
    """
    def __init__(self, reduction='mean', ignore_index: int = -100, weight=None):
        super(CrossEntropyLoss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index
        self.weight = weight

    def forward(self, y_pred, y_true):
        b, c, h, w = y_pred.shape
        if self.weight is None:
            self.weight = torch.ones(c, dtype=torch.float32)

        return F.cross_entropy(
            y_pred, y_true,
            weight=self.weight,
            ignore_index=self.ignore_index
        )

class DiceLoss(nn.Module):
    """
    Dice Loss:
        Measure the dissimilarity between the predicted and target binary segmentation
    masks. It is calculated as 1 minus the Dice coefficient, which is a measure of overlap between the predicted and target regions.

    Args:
        reduction (str, optional): Specify the reduction method applied to the output.
            The optional values are 'none', 'mean', or 'sum'. The default is' mean '.

    Returns:
        torch.Tensor: The Dice Loss between input and target.
    """
    def __init__(self, eps=1e-5):
        super(DiceLoss, self).__init__()
        self.eps = eps

    def forward(self, y_pred, y_true):
        intersection = torch.sum(y_pred * y_true)
        union = torch.sum(y_pred) + torch.sum(y_true)
        dice_loss = 1 - (2 * intersection + self.eps) / (union + self.eps)
        return dice_loss.mean()


class FocalLoss(nn.Module):
    """
    Focal Loss:
        Used to solve the problem of class imbalance, focusing on difficult to classify
    categories by reducing the loss of easy to classify categories。

    Args:
        alpha (float, optional): Control the weight of easy to classify categories, with a value
            greater than 1 indicating an increase in weight and a value less than 1 indicating a decrease in weight. The default is 1
        gamma (float, optional): Control the rate of loss reduction for difficult to classify categories, where a value greater than 0
            indicates a slower decrease and a value less than 0 indicates a faster decrease. The default is 2
        reduction (str, optional): Specify the reduction method applied to the output.
            The optional values are 'none', 'mean', or 'sum'. The default is' mean '.
    """
    def __init__(self,
                 alpha: float = .5,
                 gamma: float = 2,
                 ignore_index: int = -100,
                 reduction: str = 'none',
                 ):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, y_pred, y_true) -> torch.Tensor:
        logpt = -nn.CrossEntropyLoss(weight=None, ignore_index=self.ignore_index,
                                     reduction=self.reduction)(y_pred, y_true)
        pt = torch.exp(logpt)
        logpt *= self.alpha
        loss = -((1 - pt) ** self.gamma) * logpt
        return loss.mean()


class JaccardLoss(nn.Module):
    """
    Jaccard Loss:
        Used to measure the similarity between predicted segmentation and target segmentation.
    It is measured by calculating the ratio between the intersection and union of the predicted
    area and the target area.

    Args:
        eps (float, optional): Avoid small values divided by zero. The default is 1e-5.
    """
    def __init__(self, eps=1e-5):
        super(JaccardLoss, self).__init__()
        self.eps = eps

    def forward(self, y_pred, y_true):
        intersection = torch.sum(y_pred * y_true)
        union = torch.sum(y_pred) + torch.sum(y_true)
        jaccard_loss = 1.0 - intersection / (union - intersection + self.eps)
        return jaccard_loss.mean()


class BoundaryLoss(nn.Module):
    """
    BoundaryLoss:
        Used for calculating the boundary loss of multi class segmentation, it calculates the
    boundary loss of a certain category, but it is still recommended for binary classification

    Returns:
        torch.Tensor: The Boundary Loss between model predictions and ground truth.
    """
    def __init__(self, selected_class=1):
        super(BoundaryLoss, self).__init__()
        self.selected_class = selected_class

    def forward(self, y_pred, y_true):
        """
        compute boundary loss for multi-class segmentation
        Args:
            y_pred: model outputs, shape=(b,c,x,y)
            y_true: one-hot encoded ground truth, shape=(b,c,x,y)

        Returns: boundary_loss
        """
        y_pred = torch.sigmoid(y_pred)
        # Select the specified class for boundary loss calculation
        y_true_selected = y_true[:, self.selected_class, :, :].unsqueeze(1)
        y_pred_selected = y_pred[:, self.selected_class, :, :].unsqueeze(1)
        gt_sdf_npy = self.compute_sdf(y_true_selected.cpu().numpy(), y_pred_selected.shape)
        gt_sdf = torch.from_numpy(gt_sdf_npy).float()
        pc = y_pred.type(torch.float32)
        dc = gt_sdf.type(torch.float32).to(y_pred.device)
        multipled = torch.einsum("bchw, bchw->bchw", pc, dc)
        loss = multipled.mean(dim=(1, 2, 3))
        return loss.mean()

    def compute_sdf(self, img_gt, out_shape):
        """
        compute the normalized signed distance map of binary mask
        input: segmentation, shape = (batch_size, c, x, y)
        output: the Signed Distance Map (SDM)
        sdf(x) = 0; x in segmentation boundary
                 -inf|x-y|; x in segmentation
                 +inf|x-y|; x out of segmentation
        normalize sdf to [-1, 1]
        """
        img_gt = img_gt.astype(np.uint8)
        normalized_sdf = np.zeros(out_shape)
        for b in range(out_shape[0]):  # batch size
            # ignore background
            posmask = img_gt[b].astype(bool)
            for c in range(out_shape[1]):
                if posmask.any():
                    negmask = ~posmask
                    posdis = distance_transform_edt(posmask)
                    negdis = distance_transform_edt(negmask)
                    boundary = find_boundaries(posmask, mode='inner').astype(np.uint8)
                    sdf = (negdis - np.min(negdis)) / (np.max(negdis) - np.min(negdis)) - (posdis - np.min(posdis)) / (
                            np.max(posdis) - np.min(posdis))
                    # sdf = negdis - posdis
                    sdf[boundary == 1] = 0
                    normalized_sdf[b][c] = sdf

        return normalized_sdf


if __name__ == "__main__":
    input_data = torch.Tensor([2, 3, 4, 5])
    target_data = torch.Tensor([4, 5, 6, 7])
    loss1 = nn.L1Loss()(input_data, target_data)
    loss2 = L1Loss()(input_data, target_data)
    print(loss1, loss2)
    loss1 = nn.MSELoss()(input_data, target_data)
    loss2 = L2Loss()(input_data, target_data)
    print(loss1, loss2)
    input_data = torch.randn((5,))
    target_data = torch.randint(0, 2, (5,), dtype=torch.float32)
    loss3 = nn.BCELoss()(torch.sigmoid(input_data), target_data)
    loss4 = BCELoss()(input_data, target_data)
    print(loss3, loss4)
    y_pred = torch.rand(4, 3, 5, 5)
    y_true = torch.randint(0, 2, (4, 3, 5, 5)).float()
    loss5 = MCCLoss()(y_pred, y_true)
    print(loss5)
    loss6 = nn.CrossEntropyLoss()(y_pred, y_true)
    loss7 = CrossEntropyLoss()(y_pred, y_true)
    print(loss6, loss7)
    print(y_pred.shape, y_true.shape)
    loss8 = DiceLoss()(y_pred, y_true)
    loss9 = FocalLoss()(y_pred, y_true)
    loss10 = JaccardLoss()(y_pred, y_true)
    print(loss8, loss9, loss10)

    outputs_soft = torch.rand((2, 2, 224, 224))  # model prediction
    print(torch.unique(outputs_soft))
    label_batch = torch.randint(2, (2, 2, 224, 224))  # binary segmentation mask
    print(torch.unique(label_batch))
    loss = BoundaryLoss()(outputs_soft, label_batch)
    print("Boundary Loss:", loss)


import os
import torch
from collections import deque
import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import scipy.signal as signal
from torch.utils.tensorboard import SummaryWriter
from pyzjr.data.utils.listfun import natsorted

__all__ = ["AverageMeter", "DataStatsCalculator", "SmoothedValue", "EvalMetrics", "LossHistory", "ProcessMonitor"]

class AverageMeter(object):
    """A simple class that maintains the running average of a quantity

    Example:
    ```
        loss_avg = AverageMeter()
        loss_avg.update(2)
        loss_avg.update(4)
        loss_avg() = 3
    ```
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

class DataStatsCalculator:
    def __init__(self):
        self.data = {}
        self.total = 0.0
        self.count = 0
        self.max_val = -float('inf')
        self.min_val = float('inf')
        self.max_ids = []
        self.min_ids = []

    def update(self, identifier, value, verbose=False):
        """Update statistics with new data point"""
        self.data[identifier] = value
        self.total += value
        self.count += 1

        # Update max value and identifiers
        if value > self.max_val:
            self.max_val = value
            self.max_ids = [identifier]
        elif value == self.max_val:
            self.max_ids.append(identifier)

        # Update min value and identifiers
        if value < self.min_val:
            self.min_val = value
            self.min_ids = [identifier]
        elif value == self.min_val:
            self.min_ids.append(identifier)

        if verbose:
            print(f"Added: {identifier} = {value}")

    def calculate(self):
        """Calculate and return statistics"""
        if self.count == 0:
            return None

        avg = self.total / self.count
        return {
            'average': avg,
            'max_value': self.max_val,
            'max_identifiers': self.max_ids,
            'min_value': self.min_val,
            'min_identifiers': self.min_ids,
            'count': self.count
        }

    def save(self, filename='statistics.txt'):
        """Save results to file"""
        stats = self.calculate()
        if stats is None:
            print("No valid data to save!")
            return
        sorted_keys = natsorted(self.data.keys())
        with open(filename, 'w') as f:
            f.write("ID\tValue\n")
            for key in sorted_keys:
                f.write(f"{key}\t{self.data[key]:.4f}\n")

            f.write("\nStatistics:\n")
            f.write(f"Average: {stats['average']:.4f}\n")
            f.write(f"Max: {stats['max_value']:.4f} (IDs: {', '.join(stats['max_identifiers'])})\n")
            f.write(f"Min: {stats['min_value']:.4f} (IDs: {', '.join(stats['min_identifiers'])})\n")
            f.write(f"Total entries: {stats['count']}\n")

class SmoothedValue(object):
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """
    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{value:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            max=self.max,
            value=self.value)


class EvalMetrics(object):
    """
    Segmentation evaluation metrics (IoU, Accuracy, etc.)
    For details:https://blog.csdn.net/m0_62919535/article/details/132893016
    """
    def __init__(self, num_classes, eps=1e-5):
        self.num_classes = num_classes
        self.hist = np.zeros((num_classes, num_classes), dtype=np.int64)
        self.eps = eps

    def _fast_hist(self, label_pred, label_true, num_classes):
        mask = (label_true >= 0) & (label_true < num_classes)
        hist = np.bincount(
            num_classes * label_true[mask].astype(int) +
            label_pred[mask], minlength=num_classes ** 2).reshape(num_classes, num_classes)
        return hist

    def update(self, pred, true):
        if isinstance(pred, torch.Tensor):
            pred = pred.cpu().numpy()
        if isinstance(true, torch.Tensor):
            true = true.cpu().numpy()
        for lp, lt in zip(pred, true):
            self.hist += self._fast_hist(lp.flatten(), lt.flatten(), self.num_classes)

    def compute(self, hist):
        acc = np.diag(hist).sum() / (hist.sum() + self.eps)
        acc_cls = np.diag(hist) / (hist.sum(axis=1) + self.eps)
        acc_cls = np.nanmean(acc_cls)
        iu = np.diag(hist) / (hist.sum(axis=1) + hist.sum(axis=0) - np.diag(hist) + self.eps)
        mean_iu = np.nanmean(iu)
        freq = hist.sum(axis=1) / hist.sum()
        fwavacc = (freq[freq > 0] * iu[freq > 0]).sum()
        return acc, acc_cls, mean_iu, fwavacc

    def reset(self):
        """Reset all accumulated state"""
        self.hist = np.zeros((self.num_classes, self.num_classes), dtype=np.int64)

    def eval(self):
        """Print formatted evaluation metrics"""
        acc, acc_cls, mean_iu, fwavacc = self.compute(self.hist)
        print(
            f'Global Accuracy: {acc:.2%}\n'
            f'Class Accuracy: {acc_cls:.2%}\n'
            f'Mean IoU: {mean_iu:.2%}\n'
            f'Weighted IoU: {fwavacc:.2%}'
        )

class LossHistory():
    def __init__(self, log_dir):
        self.log_dir = log_dir
        self.losses = []
        self.val_loss = []
        os.makedirs(self.log_dir, exist_ok=True)

    def append_loss(self, epoch, loss, val_loss):
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        self.losses.append(loss)
        self.val_loss.append(val_loss)

        with open(os.path.join(self.log_dir, "epoch_loss.txt"), 'a') as f:
            f.write(f"{epoch} train: {loss} val: {val_loss}")
            f.write("\n")
        self.loss_plot()

    def loss_plot(self):
        iters = range(len(self.losses))
        losses_cpu = torch.tensor(self.losses).cpu().numpy()
        val_loss_cpu = torch.tensor(self.val_loss).cpu().numpy()

        plt.figure()
        plt.plot(iters, losses_cpu, 'red', linewidth=2, label='train loss')
        plt.plot(iters, val_loss_cpu, 'coral', linewidth=2, label='val loss')

        try:
            if len(self.losses) < 25:
                num = 5
            else:
                num = 15
            smoothed_train_loss = signal.savgol_filter(losses_cpu, num, 3)
            smoothed_val_loss = signal.savgol_filter(val_loss_cpu, num, 3)

            plt.plot(iters, smoothed_train_loss, 'green', linestyle='--', linewidth=2, label='smooth train loss')
            plt.plot(iters, smoothed_val_loss, '#8B4513', linestyle='--', linewidth=2, label='smooth val loss')
        except:
            pass

        plt.grid(True)
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.legend(loc="upper right")
        plt.savefig(os.path.join(self.log_dir, "epoch_loss.png"))

        plt.cla()
        plt.close("all")


class ProcessMonitor:
    def __init__(self, epochs, metric='train_loss', mode='min',
                 save_path='./logs', figsize=(8, 6)):
        self.figsize = figsize
        self.save_path = os.path.join(save_path, f'{metric}.png')
        self.metric_name = metric
        self.metric_mode = mode
        self.epochs = epochs
        self.history = {}
        self.step, self.epoch = 0, 0

        self.csv_filename = os.path.join(save_path, f'{metric}.csv')
        os.makedirs(save_path, exist_ok=True)
        self._init_csv()

    def _init_csv(self):
        """初始化 CSV 文件，如果文件不存在，则创建一个包含列标题的空文件"""
        if not os.path.exists(self.csv_filename):
            columns = ['epoch', self.metric_name, 'train_loss', 'val_loss']
            df = pd.DataFrame(columns=columns)
            df.to_csv(self.csv_filename, index=False)

    def start(self):
        print('\nView dynamic loss/metric plot: \n' + os.path.abspath(self.save_path))
        x_bounds = [0, min(10, self.epochs)]
        title = f'best {self.metric_name} = ?'
        self.update_graph(title=title, x_bounds=x_bounds)

    def log_epoch(self, info):
        self.epoch += 1
        info['epoch'] = self.epoch

        # 更新历史记录字典
        for name, metric in info.items():
            if name not in self.history:
                self.history[name] = []
            self.history[name].append(metric)

        # 使用 pandas 更新 CSV 文件
        self._save_to_csv()

        dfhistory = pd.DataFrame(self.history)
        n = len(dfhistory)
        x_bounds = [dfhistory['epoch'].min(), min(10 + (n // 10) * 10, self.epochs)]
        title = self.get_title()
        self.step, self.batchs = 0, self.step
        self.update_graph(title=title, x_bounds=x_bounds)

    def _save_to_csv(self):
        """将当前历史记录保存到 CSV 文件"""
        df = pd.DataFrame(self.history)
        df.to_csv(self.csv_filename, index=False)

    def end(self):
        title = self.get_title()
        self.update_graph(title=title)
        self.dfhistory = pd.DataFrame(self.history)
        return self.dfhistory

    def head(self):
        best_epoch, best_score = self.get_best_score()
        print(f"Best {self.metric_name}: {best_score:.4f} at epoch {best_epoch}")
        return self.dfhistory.head()

    def get_best_score(self):
        dfhistory = pd.DataFrame(self.history)
        arr_scores = dfhistory[self.metric_name]
        best_score = np.max(arr_scores) if self.metric_mode == "max" else np.min(arr_scores)
        best_epoch = dfhistory.loc[arr_scores == best_score, 'epoch'].tolist()[0]
        return (best_epoch, best_score)

    def get_title(self):
        best_epoch, best_score = self.get_best_score()
        title = f'best {self.metric_name}={best_score:.4f} (@epoch {best_epoch})'
        return title

    def update_graph(self, title=None, x_bounds=None, y_bounds=None):
        if not hasattr(self, 'graph_fig'):
            self.fig, self.ax = plt.subplots(1, figsize=self.figsize)
        self.ax.clear()

        dfhistory = pd.DataFrame(self.history)
        epochs = dfhistory['epoch'] if 'epoch' in dfhistory.columns else []

        metric_name = self.metric_name.replace('val_', '').replace('train_', '')

        m1 = "train_" + metric_name
        if m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            self.ax.plot(epochs, train_metrics, 'bo--', label=m1, clip_on=False)

        m2 = 'val_' + metric_name
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.ax.plot(epochs, val_metrics, 'co-', label=m2, clip_on=False)

        if metric_name in dfhistory.columns:
            metric_values = dfhistory[metric_name]
            self.ax.plot(epochs, metric_values, 'co-', label=self.metric_name, clip_on=False)

        self.ax.set_xlabel("epoch")
        self.ax.set_ylabel(metric_name)

        if title:
            self.ax.set_title(title)

        if m1 in dfhistory.columns or m2 in dfhistory.columns or self.metric_name in dfhistory.columns:
            self.ax.legend(loc='best')

        if len(epochs) > 0:
            best_epoch, best_score = self.get_best_score()
            self.ax.plot(best_epoch, best_score, 'r*', markersize=15, clip_on=False)

        if x_bounds is not None: self.ax.set_xlim(*x_bounds)
        if y_bounds is not None: self.ax.set_ylim(*y_bounds)
        self.fig.savefig(self.save_path)
        plt.close()




if __name__=='__main__':
    # import random
    # epochs = 20
    # vlog = ProcessMonitor(epochs=epochs, metric='train_loss', mode='min')
    # vlog.start()
    # for epoch in range(1, epochs + 1):
    #     train_loss = random.uniform(0.5, 2.0)
    #     val_loss = random.uniform(0.4, 1.5)
    #     vlog.log_epoch({'train_loss': train_loss, 'val_loss': val_loss})
    #
    # vlog.end()
    # print(vlog.head())
    #
    #
    # vlog = ProcessMonitor(epochs=epochs, metric='psnr', mode='max')
    # for epoch in range(1, epochs + 1):
    #     psnr = random.uniform(25, 42)
    #     vlog.log_epoch({'psnr': psnr})
    #
    # vlog.end()
    # print(vlog.head())

    true = torch.tensor([0, 1, 2, 0, 3, 2, 4])
    pred = torch.tensor([0, 1, 1, 0, 3, 1, 4])  # 第三个样本预测错误
    cm = EvalMetrics(num_classes=5)
    cm.update(true, pred)
    cm.eval()









import cv2
import math
import numpy as np
import pyzjr.Z as Z
from PIL import Image, ImageDraw


def DrawPolygon(img, points, color):
    """
    Draw a polygon onto an image using the given points and fill color.
    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) on which the rectangle and corners will be drawn.
    points : list of tuples or numpy.ndarray
            List of vertex coordinates for polygons. Each coordinate should be a tuple (x, y) containing two integers,
            Alternatively, a NumPy array with the shape (n, 1, 2) can be passed, where n is the number of vertices of the polygon.
    color : tuple
        Fill the color of the polygon in the format of (B, G, R), with each value ranging from 0 to 255 (inclusive)
    Code Example:
    ------------
    >>> image = np.ones((400, 400, 3), dtype=np.uint8)
    >>> polygon_points = np.array([[100, 100], [200, 50], [300, 150], [250, 300], [150, 250]])
    >>> result_image = DrawPolygon(image, polygon_points,  (255, 0, 0))
    """
    points = np.array([points])
    points = points.astype(np.int32)
    img = cv2.fillPoly(img, points, color)
    return img


def DrawCornerRectangle(img, bbox, length=30, lthickness=5, rthickness=1,
                        bboxcolor=Z.purple, cornercolor=Z.green):
    """
    Draws a rectangle with highlighted corners on an image. The rectangle is defined by a bounding box,
    and the corners are emphasized with lines of a specified length and thickness.

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) on which the rectangle and corners will be drawn.
    bbox : tuple
        The bounding box coordinates in the format `(x, y, x1, y1)`, where `(x, y)` is the top-left corner
        and `(x1, y1)` is the bottom-right corner.
    length : int, optional
        The length of the corner lines (default is 30 pixels).
    lthickness : int, optional
        The thickness of the corner lines (default is 5).
    rthickness : int, optional
        The thickness of the rectangle lines (default is 1).
    bboxcolor : tuple, optional
        The color of the rectangle in BGR format (e.g., (128, 0, 128) for purple). Default is `Z.purple`.
    cornercolor : tuple, optional
        The color of the corner lines in BGR format (e.g., (0, 255, 0) for green). Default is `Z.green`.

    Returns:
    --------
    numpy.ndarray
        The modified image with the rectangle and corners drawn.

    Usage Example:
    --------------
    >>> img = np.zeros((300, 500, 3), dtype=np.uint8)
    >>> bbox = (50, 50, 200, 150)
    >>> img = DrawCornerRectangle(img, bbox, length=30, lthickness=5, rthickness=1,
    ...                          bboxcolor=(128, 0, 128), cornercolor=(0, 255, 0))
    """
    x, y, x1, y1 = bbox
    cv2.rectangle(img, (x, y), (x1, y1), bboxcolor, rthickness)
    # Top-left corner (x, y)
    cv2.line(img, (x, y), (x + length, y), cornercolor, lthickness)
    cv2.line(img, (x, y), (x, y + length), cornercolor, lthickness)
    # Top-right corner (x1, y)
    cv2.line(img, (x1, y), (x1 - length, y), cornercolor, lthickness)
    cv2.line(img, (x1, y), (x1, y + length), cornercolor, lthickness)
    # Bottom-left corner (x, y1)
    cv2.line(img, (x, y1), (x + length, y1), cornercolor, lthickness)
    cv2.line(img, (x, y1), (x, y1 - length), cornercolor, lthickness)
    # Bottom-right corner (x1, y1)
    cv2.line(img, (x1, y1), (x1 - length, y1), cornercolor, lthickness)
    cv2.line(img, (x1, y1), (x1, y1 - length), cornercolor, lthickness)

    return img


def select_roi_region(image_path, line_color=(0, 255, 0), zoom_factor=3):
    """
    Interactive ROI selection tool with real-time zoom preview
    Args:
        image_path: Path to input image
        line_color: BGR color tuple for rectangle (default: green)
        zoom_factor: Multiplier for preview window zoom (default: 3x)
    Returns:
        tuple: (x, y, width, height) of selected region
    """
    drawing = False
    ix, iy = -1, -1
    x, y, w, h = 0, 0, 0, 0
    img = cv2.imread(image_path)
    clone = img.copy()

    # 鼠标回调函数
    def mouse_callback(event, cur_x, cur_y, flags, param):
        nonlocal ix, iy, drawing, x, y, w, h

        if event == cv2.EVENT_LBUTTONDOWN:
            drawing = True
            ix, iy = cur_x, cur_y
            x, y, w, h = 0, 0, 0, 0

        elif event == cv2.EVENT_MOUSEMOVE and drawing:
            temp_img = clone.copy()
            cv2.rectangle(temp_img, (ix, iy), (cur_x, cur_y), line_color, 2)
            x1, y1 = min(ix, cur_x), min(iy, cur_y)
            x2, y2 = max(ix, cur_x), max(iy, cur_y)

            if x2 > x1 and y2 > y1:
                try:
                    roi = img[y1:y2, x1:x2]
                    if roi.size > 0:
                        enlarged = cv2.resize(roi, None, fx=3, fy=3,
                                              interpolation=cv2.INTER_CUBIC)
                        cv2.imshow("Enlarged Preview", enlarged)
                except Exception as e:
                    pass

            cur_w = abs(cur_x - ix)
            cur_h = abs(cur_y - iy)
            if cur_w > 0 and cur_h > 0:
                try:
                    roi = img[y1:y2, x1:x2]
                    enlarged = cv2.resize(roi, None, fx=zoom_factor, fy=zoom_factor,
                                          interpolation=cv2.INTER_CUBIC)
                    cv2.imshow("Enlarged Preview", enlarged)
                except:
                    pass
            cv2.putText(temp_img, f"X:{x1} Y:{y1} W:{cur_w} H:{cur_h}",
                        (10, 30), cv2.FONT_HERSHEY_SIMPLEX, 0.7, line_color, 2)
            cv2.imshow("Select ROI (SPACE=Clear | ENTER=Confirm)", temp_img)

        elif event == cv2.EVENT_LBUTTONUP:
            drawing = False
            x = min(ix, cur_x)
            y = min(iy, cur_y)
            w = abs(cur_x - ix)
            h = abs(cur_y - iy)
            cv2.rectangle(clone, (x, y), (x + w, y + h), line_color, 2)
            cv2.putText(clone, f"X:{x} Y:{y} W:{w} H:{h}", (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.7, line_color, 2)
            cv2.imshow("Select ROI (SPACE=Clear | ENTER=Confirm)", clone)

    cv2.namedWindow("Select ROI (SPACE=Clear | ENTER=Confirm)")
    cv2.setMouseCallback("Select ROI (SPACE=Clear | ENTER=Confirm)", mouse_callback)

    while True:
        cv2.imshow("Select ROI (SPACE=Clear | ENTER=Confirm)", clone)
        key = cv2.waitKey(1) & 0xFF
        # 空格键：清除选择
        if key == 32:
            clone = img.copy()
            ix, iy = -1, -1
            x, y, w, h = 0, 0, 0, 0
            try:
                cv2.destroyWindow("Enlarged Preview") if cv2.getWindowProperty("Enlarged Preview", 0) >= 0 else None
            except:
                pass
            cv2.imshow("Select ROI (SPACE=Clear | ENTER=Confirm)", clone)
        # 回车键：确认选择
        if key == 13:
            try:
                cv2.destroyWindow("Enlarged Preview")
            except:
                pass
            break
    cv2.destroyAllWindows()
    print(f"Final selection - X:{x} Y:{y} W:{w} H:{h}")
    return (x, y, w, h)


def plot_highlight_region(image_path, region_to_zoom, paste_position=None, zoom_factor=3,
                          line_color="red", line_wide=2, show_arrow=True, arrow_size=5):
    """
    Visualize image regions with zoomed preview and connection arrows.

    Args:
        image_path: Path to source image
        region_to_zoom: Tuple (x, y, w, h) defining region to highlight
        paste_position: Custom position (x,y) for zoomed preview (auto-placed if None)
        zoom_factor: Scaling multiplier for preview
        line_color: Border/arrow color name or hex code
        line_wide: Border/arrow line width
        show_arrow: Toggle connection arrow visibility
        arrow_size: Relative size multiplier for arrowhead

    Returns:
        PIL.Image: Composite image with visual annotations
   """
    x, y, w, h = region_to_zoom
    img = Image.open(image_path).convert("RGB")
    img_w, img_h = img.size
    original_copy = img.copy()
    zoomed_w = int(w * zoom_factor)
    zoomed_h = int(h * zoom_factor)
    cropped = original_copy.crop((x, y, x + w, y + h))
    zoomed = cropped.resize((zoomed_w, zoomed_h), Image.Resampling.LANCZOS)
    if paste_position is None:
        if x + w < img_w / 2:
            paste_x = img_w - zoomed_w
        else:
            paste_x = 0
        if y + h < img_h / 2:
            paste_y = img_h - zoomed_h
        else:
            paste_y = 0
        paste_x = max(0, min(paste_x, img_w - zoomed_w))
        paste_y = max(0, min(paste_y, img_h - zoomed_h))
        paste_position = (paste_x, paste_y)
    img.paste(zoomed, paste_position)
    draw = ImageDraw.Draw(img)
    draw.rectangle([(x, y), (x + w, y + h)],
                   outline=line_color,
                   width=line_wide)
    paste_x, paste_y = paste_position
    draw.rectangle([paste_position,
                    (paste_x + zoomed_w, paste_y + zoomed_h)],
                   outline=line_color, width=line_wide)
    if show_arrow:
        def get_side_center(rect, side):
            x, y, w, h = rect
            return {
                'left': (x, y + h // 2),
                'right': (x + w, y + h // 2),
                'top': (x + w // 2, y),
                'bottom': (x + w // 2, y + h)
            }[side]

        src_rect = (x, y, w, h)
        dst_rect = (paste_position[0], paste_position[1], zoomed_w, zoomed_h)
        dx = (dst_rect[0] + zoomed_w / 2) - (x + w / 2)
        dy = (dst_rect[1] + zoomed_h / 2) - (y + h / 2)
        if abs(dx) > abs(dy):
            src_side = 'right' if dx > 0 else 'left'
            dst_side = 'left' if dx > 0 else 'right'
        else:
            src_side = 'bottom' if dy > 0 else 'top'
            dst_side = 'top' if dy > 0 else 'bottom'

        start_point = get_side_center(src_rect, src_side)
        end_point = get_side_center(dst_rect, dst_side)
        draw.line([start_point, end_point], fill=line_color, width=line_wide)
        arrow_size = line_wide * arrow_size
        angle = math.atan2(end_point[1] - start_point[1], end_point[0] - start_point[0])
        p1 = (end_point[0] - arrow_size * math.cos(angle - math.pi / 6),
              end_point[1] - arrow_size * math.sin(angle - math.pi / 6))
        p2 = (end_point[0] - arrow_size * math.cos(angle + math.pi / 6),
              end_point[1] - arrow_size * math.sin(angle + math.pi / 6))
        draw.polygon([end_point, p1, p2], fill=line_color)
    return img


if __name__ == "__main__":
    imagePath = r"E:\PythonProject\pyzjrPyPi\pyzjr\utils\tryout\images\cat.png"
    image = np.ones((400, 400, 3), dtype=np.uint8)
    polygon_points = np.array([[100, 100], [200, 50], [300, 150], [250, 300], [150, 250]])
    result_image = DrawPolygon(image, polygon_points, (255, 0, 0))
    img = cv2.imread(imagePath)
    # img = np.zeros((300, 500, 3), dtype=np.uint8)
    bbox = (50, 50, 200, 150)
    img = DrawCornerRectangle(img, bbox, length=30, lthickness=5, rthickness=1,
                              bboxcolor=(128, 0, 128), cornercolor=(0, 255, 0))
    cv2.imshow("show", img)
    cv2.imshow("show2", result_image)
    cv2.waitKey(0)

    im_path = r'E:\PythonProject\img_processing_techniques_main\Enlarge_local_details\gtimage\861.png'
    region_to_zoom = select_roi_region(im_path)
    im = plot_highlight_region(im_path, region_to_zoom, show_arrow=False)
    # im.save("output.png")



"""
Copyright (c) 2022, Auorui.
All rights reserved.

Call torchvision for image enhancement
"""
import random
import torch
import torchvision
import numpy as np
import PIL.Image as Image
from pyzjr.utils.randfun import rand
import torchvision.transforms.functional as tf
from pyzjr.augmentation.transforms.apply import Images

def pad_if_smaller(img, size, fill=0):
    """If the minimum edge length of the image is smaller than the given size, use a numerical fill for padding"""
    min_size = min(img.size)
    if min_size < size:
        ow, oh = img.size
        padh = size - oh if oh < size else 0
        padw = size - ow if ow < size else 0
        img = tf.pad(img, [0, 0, padw, padh], fill=fill)
    return img

class tvisionToTensor(object):
    def __call__(self, image, target):
        image = tf.to_tensor(image)
        target = torch.as_tensor(np.array(target), dtype=torch.int64).permute(2, 0, 1)
        return image, target

class tvisionRandomResize(object):
    def __init__(self, min_size, max_size=None):
        self.min_size = min_size
        if max_size is None:
            max_size = min_size
        self.max_size = max_size

    def __call__(self, image, target):
        size = random.randint(self.min_size, self.max_size)
        # The size input here is of type int, so the minimum edge length of the image is scaled to size
        image = tf.resize(image, [size])
        # Note that the interpolation here is only available after torchvision (0.9.0) in InterpolationMode. NEAREST
        # If it is a previous version, PIL.ImageNEAREST needs to be used
        if torchvision.__version__>= "0.9.0":
            NEAREST = torchvision.transforms.InterpolationMode.NEAREST
        else:
            NEAREST = Images.NEAREST
        target = tf.resize(target, [size], interpolation=NEAREST)
        return image, target

class tvisionRandomHorizontalFlip(object):
    def __init__(self, prob):
        self.prob = prob

    def __call__(self, image, target):
        if random.random() < self.prob:
            image = tf.hflip(image)
            target = tf.hflip(target)
        return image, target

class tvisionRandomVerticalFlip(object):
    def __init__(self, prob):
        self.prob = prob

    def __call__(self, image, target):
        if random.random() < self.prob:
            image = tf.vflip(image)
            target = tf.vflip(target)
        return image, target

class tvisionRandomCrop(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = pad_if_smaller(image, self.size)
        target = pad_if_smaller(target, self.size, fill=0)
        crop_params = torchvision.transforms.RandomCrop.get_params(image, (self.size, self.size))
        image = tf.crop(image, *crop_params)
        target = tf.crop(target, *crop_params)
        return image, target

class tvisionCenterCrop(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = tf.center_crop(image, self.size)
        target = tf.center_crop(target, self.size)
        return image, target

class tvisionNormalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, target):
        image = tf.normalize(image, mean=self.mean, std=self.std)
        return image, target

class tvisionRandomBrightness(object):
    def __init__(self, brightness_range=(0.5, 2.0)):
        self.brightness_range = brightness_range

    def __call__(self, image, target):
        brightness_factor = rand(self.brightness_range[0], self.brightness_range[1])
        image = tf.adjust_brightness(image, brightness_factor)
        return image, target

class tvisionRandomContrast(object):
    def __init__(self, contrast_range=(0.5, 2.0)):
        self.contrast_range = contrast_range

    def __call__(self, image, target):
        contrast_factor = rand(self.contrast_range[0], self.contrast_range[1])
        image = tf.adjust_contrast(image, contrast_factor)
        return image, target

class tvisionPerspectiveTransform:
    def __init__(self, distortion_scale=0.5, p=0.5, fill=None):
        self.distortion_scale = distortion_scale
        self.p = p
        self.fill = fill

    def __call__(self, img, target):
        if random.random() < self.p:
            width, height = img.size
            half_h = height // 2
            half_w = width // 2
            topleft = [
                random.randint(0, int(self.distortion_scale * half_w)),
                random.randint(0, int(self.distortion_scale * half_h))
            ]
            topright = [
                random.randint(width - int(self.distortion_scale * half_w), width - 1),
                random.randint(0, int(self.distortion_scale * half_h))
            ]
            botright = [
                random.randint(width - int(self.distortion_scale * half_w), width - 1),
                random.randint(height - int(self.distortion_scale * half_h), height - 1)
            ]
            botleft = [
                random.randint(0, int(self.distortion_scale * half_w)),
                random.randint(height - int(self.distortion_scale * half_h), height - 1)
            ]
            startpoints = [[0, 0], [width - 1, 0], [width - 1, height - 1], [0, height - 1]]
            endpoints = [topleft, topright, botright, botleft]
            img = tf.perspective(img, startpoints, endpoints,
                                 interpolation=tf.InterpolationMode.BILINEAR, fill=self.fill)
            target = tf.perspective(target, startpoints, endpoints,
                                    interpolation=tf.InterpolationMode.NEAREST, fill=self.fill)
        return img, target

class tvisionColorJitter:
    def __init__(self, brightness=0.3, contrast=0.3, saturation=0.3, hue=0.2):
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue

    def __call__(self, image, target):
        if self.brightness > 0:
            brightness_factor = rand(1-self.brightness, 1+self.brightness)
            image = tf.adjust_brightness(image, brightness_factor)
        if self.contrast > 0:
            contrast_factor = rand(1-self.contrast, 1+self.contrast)
            image = tf.adjust_contrast(image, contrast_factor)
        if self.saturation > 0:
            saturation_factor = rand(1-self.saturation, 1+self.saturation)
            image = tf.adjust_saturation(image, saturation_factor)
        if self.hue > 0:
            hue_factor = rand(-self.hue, self.hue)
            image = tf.adjust_hue(image, hue_factor)

        return image, target

class tvisionResizePad:
    def __init__(self, size, fill=128, label_fill=0):
        self.size = (size, size) if isinstance(size, int) else size
        self.fill = (fill, fill, fill) if isinstance(fill, int) else fill
        self.label_fill = (label_fill, label_fill, label_fill) if isinstance(label_fill, int) else label_fill

    def __call__(self, image, target):
        iw, ih = image.size
        w, h = self.size
        scale = min(w/iw, h/ih)
        nw = int(iw*scale)
        nh = int(ih*scale)
        image = image.resize((nw,nh), Images.CUBIC)
        new_image = Image.new('RGB', [w, h], self.fill)
        new_image.paste(image, ((w-nw)//2, (h-nh)//2))
        target = target.resize((nw,nh), Images.NEAREST)
        new_target = Image.new('RGB', [w, h], self.label_fill)
        new_target.paste(target, ((w-nw)//2, (h-nh)//2))
        return new_image, new_target


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from pyzjr.augmentation.transforms.apply import Compose
    from pyzjr.visualize.io.matplot import matplotlib_patch
    matplotlib_patch()
    image_path = r'E:\PythonProject\Pytorch_Segmentation_Auorui\data\VOCdevkit\VOC2012\JPEGImages\2007_000033.jpg'
    mask_path = r'E:\PythonProject\Pytorch_Segmentation_Auorui\data\VOCdevkit\VOC2012\SegmentationClass\2007_000033.png'

    # 创建数据增强操作的组合
    transforms = Compose([
        # tvisionRandomResize(min_size=256, max_size=512),
        # tvisionRandomHorizontalFlip(prob=0.5),
        # tvisionRandomCrop(size=256),
        # tvisionCenterCrop(size=200),
        # tvisionColorJitter(),
        tvisionResizePad(size=256),
        # tvisionPerspectiveTransform(0.6),
        tvisionToTensor(),
        # tvisionRandomContrast(),
        # tvisionRandomBrightness(),
        # tvisionNormalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])
    img = Image.open(image_path).convert('RGB')
    mask = Image.open(mask_path).convert('RGB')
    trans_img, trans_mask = transforms(img, mask)
    alpha = 0.5
    fused = alpha * trans_img.permute(1, 2, 0) + (1 - alpha) * trans_mask.permute(1, 2, 0)
    plt.figure(figsize=(12, 6))
    plt.subplot(1, 3, 1)
    plt.title("Transformed Image")
    plt.imshow(trans_img.permute(1, 2, 0))
    plt.axis('off')

    plt.subplot(1, 3, 2)
    plt.title("Transformed Mask")
    plt.imshow(trans_mask.permute(1, 2, 0))
    plt.axis('off')

    plt.subplot(1, 3, 3)
    plt.title("Fused Image & Mask")
    plt.imshow(fused)
    plt.axis('off')

    plt.tight_layout()
    plt.show()



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional, Callable, Union

class ConvNormActivation(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, tuple],
            stride: Union[int, tuple] = 1,
            padding: Union[int, tuple, str] = 0,
            dilation: Union[int, tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            apply_act: bool = True,
            conv_layer: Callable[..., nn.Module] = nn.Conv2d,
            norm_layer: Optional[Callable[..., nn.Module]] = nn.BatchNorm2d,
            activation_layer: Optional[Callable[..., nn.Module]] = nn.ReLU,
            **kwargs
    ):
        super(ConvNormActivation, self).__init__()
        self.conv = conv_layer(
            in_channels, out_channels, kernel_size,
            stride=stride, padding=padding, dilation=dilation,
            groups=groups, bias=bias, **kwargs
        )
        layers = [self.conv]
        if self.norm is not None:
            layers.append(norm_layer(out_channels))
        if activation_layer is not None and apply_act:
            layers.append(activation_layer())
        self.block = nn.Sequential(*layers)

    @property
    def in_channels(self):
        return self.conv.in_channels

    @property
    def out_channels(self):
        return self.conv.out_channels

    def forward(self, x):
        return self.block(x)

class AutopadConv2d(nn.Conv2d):
    def __init__(self, in_channels: int, out_channels: int, kernel_size, stride=1, padding=None, dilation=1,
                 groups=1, bias=True, padding_mode='reflect', **kwargs):
        k, p, d = kernel_size, padding, dilation
        if d > 1:
            k = d * (k - 1) + 1 if isinstance(k, int) else [d * (x - 1) + 1 for x in k]  # actual kernel-size
        if p is None:
            self.p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
        super(AutopadConv2d, self).__init__(in_channels, out_channels, kernel_size, stride=stride,
                                            padding=self.p, dilation=dilation, groups=groups, bias=bias,
                                            padding_mode=padding_mode, **kwargs)

class DepthwiseConv2d(nn.Conv2d):
    # Depthwise Convolution
    def __init__(self, in_channels, kernel_size, stride=1, padding=0, bias=True, **kwargs):
        super(DepthwiseConv2d, self).__init__(in_channels, in_channels, kernel_size,
                                       stride=stride, padding=padding, bias=bias, groups=in_channels, **kwargs)

class PointwiseConv2d(nn.Conv2d):
    # Pointwise Convolution
    def __init__(self, in_channels, out_channels, bias=True, **kwargs):
        super(PointwiseConv2d, self).__init__(in_channels, out_channels, 1, stride=1, padding=0, bias=bias, **kwargs)

class DepthwiseSeparableConv2d(nn.Module):
    # Depth-wise separable convolution
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, groups=1, bias=True, **kwargs):
        super(DepthwiseSeparableConv2d, self).__init__()
        self.depthwise = DepthwiseConv2d(in_channels, kernel_size, stride, padding, bias, **kwargs)
        self.pointwise = PointwiseConv2d(in_channels, out_channels, bias, groups=groups)

    def forward(self, x):
        return self.pointwise(self.depthwise(x))

class PartialConv2d(nn.Module):
    def __init__(self, dim, n_div, kernel_size=3, forward='split_cat'):
        """
        PartialConv
        code from: https://github.com/JierunChen/FasterNet/blob/master/models/fasternet.py
        """
        super().__init__()
        self.dim_conv3 = dim // n_div
        self.dim_untouched = dim - self.dim_conv3
        self.partial_conv3 = nn.Conv2d(self.dim_conv3, self.dim_conv3, kernel_size, 1, 1, bias=False)

        if forward == 'slicing':
            self.forward = self.forward_slicing
        elif forward == 'split_cat':
            self.forward = self.forward_split_cat
        else:
            raise NotImplementedError

    def forward_slicing(self, x):
        # only for inference
        x = x.clone()   # !!! Keep the original input intact for the residual connection later
        x[:, :self.dim_conv3, :, :] = self.partial_conv3(x[:, :self.dim_conv3, :, :])

        return x

    def forward_split_cat(self, x):
        # for training/inference
        x1, x2 = torch.split(x, [self.dim_conv3, self.dim_untouched], dim=1)
        x1 = self.partial_conv3(x1)
        x = torch.cat((x1, x2), 1)
        return x


class LayerNorm(nn.Module):
    r""" LayerNorm implementation used in ConvNeXt
    LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).
    """
    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last", reshape_last_to_first=False):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)
        self.reshape_last_to_first = reshape_last_to_first

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x


import torch
import torch.nn as nn
from typing import Tuple

class JointLoss(nn.Module):
    def __init__(self, *criterions, loss_weight:Tuple[int, float, float]):
        """Weighted summation of combining multiple loss functions

        Args:
            *criterions (nn.Module): Multiple loss function modules
            loss_weight (Tuple[int, float, float]): Corresponding weights for each loss function
        """
        super().__init__()
        if len(criterions) != len(loss_weight):
            raise ValueError(f"Number of criterions ({len(criterions)}) "
                             f"must match weights count ({len(loss_weight)})")
        for i, criterion in enumerate(criterions):
            self.add_module(f'criterion_{i}', criterion)

        self.register_buffer('loss_weights',
                             torch.tensor(loss_weight, dtype=torch.float32))

    def forward(self, y_pred, y_true):
        total_loss = 0.0
        for i, (weight, criterion) in enumerate(
                zip(self.loss_weights, self.children())
        ):
            loss = criterion(y_pred, y_true)
            total_loss += weight * loss

        return total_loss

if __name__=="__main__":
    from pyzjr.nn.losses.loss_function import CrossEntropyLoss, JaccardLoss, FocalLoss
    y_pred = torch.rand(4, 3, 5, 5)
    y_true = torch.randint(0, 2, (4, 3, 5, 5)).float()
    loss11 = JointLoss(CrossEntropyLoss(), JaccardLoss(), FocalLoss(),
                       loss_weight=(1, 0.2, 0.1))(y_pred, y_true)
    print(loss11)



from .adjust_color import adjust_brightness_with_opencv, adjust_brightness_with_numpy, adjust_brightness_contrast, \
    adjust_gamma, enhance_hsv, random_hsv, hist_equalize, random_pca_lighting, color_jitter

from .blur import gaussian_blur, mean_blur, median_blur, motion_blur, bilateral_filter, guided_filter

from .binary import uint2single, single2uint, binarization, approximate, ceilfloor, up_low, bool2mask, \
    auto_canny, adaptive_bgr_threshold, create_rectmask, remove_mask_parti_color, move_mask_foreground, \
    inpaint_defect

from .contour import incircleV1, incircleV2, outcircle, check_points_in_contour, calculate_contour_lengths, \
    label_contours, sort_contours, drawOutline, SearchOutline, gradient_outline, find_contours_custom

from .dt import euclidean_dt, chamfer_dt, fast_marching_dt

from .geometric import crop_image_by_2points, crop_image_by_1points, center_crop, five_crop, stitcher_image, \
    centerzoom, flip, vertical_flip, horizontal_flip, diagonal_flip, translate, rotate, rotate_bound,\
    resize, resize_samescale, resizepad, croppad_resize, random_rotation, random_rot90, random_crop, \
    random_horizontal_flip, random_vertical_flip, pad_margin, erase, random_resize_crop, random_perspective

from .noise import salt_pepper_noise, gaussian_noise, addfog_with_channels, addfog_with_asm, add_poisson_noise, \
    add_rayleigh_noise, add_gamma_noise, add_uniform_noise

from .skeleton_extraction import medial_axis_mask, skeletonizes, thinning, read_skeleton


from .utils import OverlayPng
from .draw import DrawPolygon, DrawCornerRectangle, select_roi_region, plot_highlight_region
from .text import AddText, PutMultiLineText, PutMultiLineCenteredText, PutBoxText, PutRectangleText


from .lr_scheduler import _LRScheduler
from .adjust_lr import (
    OnceCycleLR,
    PolyLR,
    WarmUpLR,
    WarmUpWithTrainloader,
    GradualWarmupScheduler,
    CosineAnnealingLR,
    CosineAnnealingLRWithDecay,
    CosineAnnealingWarmRestartsWithDecay,
    plot_lr_scheduler,
)
from .lr_interface import get_lr, get_optimizer, get_lr_scheduler


from .check import (
    is_tensor, is_pil, is_numpy, is_gray_image, is_rgb_image, is_parallel, is_Iterable, is_str,
    is_int, is_float, is_bool, is_list, is_tuple, is_list_or_tuple, is_none, is_not_none,
    is_positive_int, is_nonnegative_int, is_ascii, is_url, is_image_extension, is_video_extension,
    is_file,is_directory, is_directory_not_empty, is_path_exists, is_windows, is_linux,
    is_odd, is_even, is_list_of_tensor, is_list_of_numpy
)

from .mathfun import (
    euclid, chessboard, cityblock, centerpoint, normal, gaussian2d, retain, round_up, round_down,
    factorial, pow, sqrt, rsqrt, cos, sin, tan, cot, sec, csc, arccos, arcsin, arctan, angle_to_2pi,
    to_degree, to_radians, exp, log_e, sinh, arsinh, cosh, arcosh, tanh, artanh, sigmoid, relu
)

from .torch_np import (
    count_zero, count_white, count_nonzero, allclose, moveaxis, in1d, clip, where, argwhere, argsort,
    nonzero, floor_divide, unravel_index, unravel_indices, ravel, any_np_pt, maximum, concatenate,
    cumsum, isfinite, searchsorted, repeat, isnan, ascontiguousarray, stack, unique, max, min, median,
    mean, std,
)

from .dim_utils import (
    pil_to_ndarray, ndarray_to_pil, to_numpy, to_tensor, to_bchw, bchw_tensor_to_bgr_image, bchw_tensor_to_list,
    bgr_image_list_to_tensor, bgr_image_to_bchw_tensor, hwc2chw, chw2hwc
)

from .randfun import (
    rand, randint, randbool, randfluctuate, randlist, randintlist, randnormal, randshuffle,
    randstep, randstring, rand_choice_weighted
)

